import os
import re
from invoke import Context, Config

U_SUCCESS = u"\u2713"
U_FAILURE = u"\u2716"

COLOR_RED = "red"
COLOR_GREEN = "green"
COLOR_YELLOW = "yellow"
COLOR_BLUE = "blue"
COLOR_PURPLE = "purple"
COLOR_LIGHTBLUE = "lightblue"
COLOR_LIGHTGREY = "lightgrey"
COLOR_GREY = "grey"
COLOR_PINK = "pink"


def get_context():
    """
    Get an invoke context object, used to run shell commands.

    :return {invoke.context.Context}: Returns the context object.
    """
    return Context(Config())


def run(cmd, ctx=None, *args, **kwargs):
    if not ctx:
        ctx = get_context()
    return ctx.run(cmd, *args, **kwargs)


def out(msg, color="green", run=True, ctx=None, is_error=False, **kwargs):
    """
    Run an echo command through invoke that prints text in color.

    We will be using the following color codes:

    Black        0;30     Dark Gray     1;30
    Red          0;31     Light Red     1;31
    Green        0;32     Light Green   1;32
    Brown/Orange 0;33     Yellow        1;33
    Blue         0;34     Light Blue    1;34
    Purple       0;35     Light Purple  1;35
    Cyan         0;36     Light Cyan    1;36
    Light Gray   0;37     White         1;37

    :param msg {str}:
    :param color {str}:
    :param run {bool}:
    :param ctx {invoke.context.Context}: Invoke context variable
    :param is_error {bool}:
        If True, echo output to stderr.
    :return {NoneType}:
    """
    if not ctx:
        ctx = get_context()

    if os.environ.get("INVOKE_ASCII", False) == "1":
        msg = msg.replace(U_SUCCESS, "PASS")
        msg = msg.replace(U_FAILURE, "FAIL")
        msg = msg.encode("utf-8").decode("ascii", "ignore")

    if run and os.environ.get("INVOKE_NO_COLOR", False) == "1":
        msg = re.sub(r"\s*\$\(tput[a-z0-9 ]+\)\s*", "", msg)
        ctx.run('echo "{}"'.format(msg), **kwargs)
        return msg

    reset = "\033[0m"
    colors = {
        COLOR_RED: "\033[0;31m",
        COLOR_GREEN: "\033[0;32m",
        COLOR_YELLOW: "\033[0;33m",
        COLOR_BLUE: "\033[0;34m",
        COLOR_PURPLE: "\033[0;35m",
        COLOR_LIGHTBLUE: "\033[1;34m",
        COLOR_LIGHTGREY: "\033[0;37m",
        COLOR_GREY: "\033[1;30m",
        COLOR_PINK: "\033[1;31m",
    }

    color = "" if (not color or color not in colors) else colors[color]
    msg = "{}{}{}".format(color, msg, reset)
    if run:
        # Output to stderr - https://stackoverflow.com/a/23550347/3330552
        prefix = ">&2 " if is_error else ""

        ctx.run('{}echo "{}"'.format(prefix, msg), **kwargs)
    return msg



from .base import BaseCommand
from odoo_manager.core import shell


class Ps(BaseCommand):
    """
    Represents the `odoo-manager ps` command which lists the current running or
    non-running containers.
    """

    def __init__(self, options, *args, **kwargs):
        """
        Inheriting the parent class for initializes command to defined that this
        command does not depend on a project.

        See the parent `__init__` method for parameters.
        """
        super(Ps, self).__init__(options, depends_on_project=False, *args, **kwargs)

    def run(self):
        """
        Runs the command. See the parent class for more details about how
        commands are organized and ran.

        :return {NoneType}:
        """
        if not self.options.get("--all", False):
            self.ctx.run("docker-compose ps")
        else:
            self.ctx.run("docker ps")



import os
from .base import BaseCommand
from odoo_manager.core import configs, shell, validation
from odoo_manager.core.git import git_clone, redirect


class Build(BaseCommand):
    def __init__(self, options, *args, **kwargs):
        """
        Inheriting the parent class for initializes command to defined that this
        command does not depend on a project.

        See the parent `__init__` method for parameters.
        """
        super(Build, self).__init__(options, depends_on_project=False, *args, **kwargs)

    # TODO: odoo-manager build[--no-cache][--no-depends][--no-setup][--pull][--verbose]
    def run(self):
        """
        This method builds your project.  It installs requirements.txt, builds
        the docker image, makes required directories and installs odoo.

        :return {NoneType}:
        """
        nocache = self.options.get("--no-cache", False)
        nodepends = self.options.get("--no-depends", False)
        nosetup = self.options.get("--no-setup", False)
        pull = self.options.get("--pull", False)
        verbose = self.options.get("--verbose", False)
        paths = self.get_paths()
        config = configs.load_standard_config(paths.base(".env"))
        user = ""
        password = ""

        validation.check_env(config, ["ODOO_VERSION", "USE_ENTERPRISE", "GIT_DEPTH"])

        shell.out("Running build process...")

        shell.out("  *Install pip3 requirements...")

        self.ctx.run("pip3 install {} --quiet -r {}".format("--user", paths.base("requirements.txt")))

        shell.out("  *Building docker image with docker-compose build...")

        self.ctx.run(
            "docker-compose build{}{}{}".format(
                " --pull" if pull else "", " --no-cache" if nocache else "", redirect(verbose)
            )
        )

        shell.out("  *Making directories...")

        if not os.path.isdir(paths.base(".container/testresults")):
            self.ctx.run("mkdir -p {}".format(paths.base(".container/testresults")))

        odoo_version = config.get("options", "ODOO_VERSION")
        use_enterprise = config.get("options", "USE_ENTERPRISE") != "0"
        depth = config.get("options", "GIT_DEPTH")

        shell.out("  *Installing odoo source...")
        shell.out("      odoo version:      {}".format(odoo_version), color="lightgrey")
        shell.out("      using enterprise?: {}".format(use_enterprise), color="lightgrey")
        shell.out("      repo depth:        {}".format(depth), color="lightgrey")
        shell.out("", color="lightgrey")

        if not os.path.isdir(paths.base(".container/core/addons")):
            if os.path.isdir(paths.base(".container/core")):
                shell.out("      Cleaning up existing core folder...")
                self.ctx.run("rm -rf {}".format(paths.base(".container/core")))
            shell.out("      Cloning down repo from github.com/odoo/odoo...")
            git_clone(
                self.ctx,
                "https://github.com/odoo/odoo",
                odoo_version,
                "odoo/odoo",
                output=paths.base(".container/core"),
                depth=depth,
            )

        us_locale_path = paths.base(".container/core/addons/web/static/lib/moment/locale/en-us.js")
        ca_locale_path = paths.base(".container/core/addons/web/static/lib/moment/locale/en-ca.js")

        if not os.path.isfile(us_locale_path) and os.path.isfile(ca_locale_path):
            shell.out("  *Cleaning up moment locales...")
            self.ctx.run("cp {} {}".format(ca_locale_path, us_locale_path))

        if use_enterprise:
            if not os.path.isdir(paths.base(".container/enterprise/account_reports")):
                if os.path.isdir(paths.base(".container/enterprise")):
                    shell.out("      Cleaning up existing enterprise folder...")
                    self.ctx.run("rm -rf {}".format(paths.base(".container/enterprise")))
                shell.out("      Cloning down repo from github.com/odoo/enterprise...")
                git_clone(
                    self.ctx,
                    "https://github.com/odoo/enterprise",
                    odoo_version,
                    "odoo/enterprise",
                    output=paths.base(".container/enterprise"),
                    depth=depth,
                    user=user,
                    password=password,
                )



from .base import BaseCommand
from odoo_manager.core import shell
from odoo_manager.core.git import redirect


class Stop(BaseCommand):
    """
    Represents the `odoo-manager stop` command which shuts down the running
    instances locally for the current odoo project.
    """

    def run(self):
        """
        Runs the command. See the parent class for more details about how
        commands are organized and ran.

        :return {NoneType}:
        """
        shell.out("Stopping containers...")
        self.ctx.run("docker-compose stop{}".format(redirect(verbose=False)))
        shell.out("Containers stopped.")



import os
from glob import glob
from odoo_manager.cli.exceptions import CommandException

pipe_dev_null = "> /dev/null 2>&1"

# This is the file pattern which indicates the project root directory.
COMPOSE_PATTERN = "docker-compose.*yml"


class Paths(object):
    base_path = None

    def __init__(self):
        """
        Initialize a Paths object.
        """
        iterated = 0
        working_dir = os.getcwd()

        while not self.base_path and iterated < 100:
            iterated += 1
            compose_filenames = glob(COMPOSE_PATTERN)
            for compose_filename in compose_filenames:
                if os.path.isfile(os.path.join(working_dir, compose_filename)):
                    self.base_path = working_dir
                    break
            working_dir = os.path.abspath(os.path.join(working_dir, ".."))

        if not self.base_path:
            raise CommandException(
                "We cannot find a valid project. Make sure that you are inside of an Odoo project with a docker-compose.yml file."
            )

    def base(self, path=""):
        """
        Helper method to ensure all paths are relative to tasks.py.

        :param path:
        :return:
        """
        return os.path.join(self.base_path, path)


def iterate_py_files(fn, base):
    """
    :param fn:
    :return {NoneType}:
    """
    ignore = [".container", ".deployment", ".docs", ".githooks", ".monitoring", ".om", "_lib", "_lib_static"]
    for dirpath, dirs, filenames in os.walk(base, topdown=True):
        dirs[:] = [d for d in dirs if d not in ignore]
        for filename in [f for f in filenames if f.endswith(".py") and f not in ignore]:
            fn(filename, os.path.join(dirpath, filename))


def iterate_odoo_modules(fn):
    """
    :param fn:
    :return {NoneType}:
    """
    pass



import os
import sys
from pprint import pprint as pp
from odoo_manager.core import shell
from odoo_manager.cli.commands.format import Format
from .base import BaseCommand

try:
    import pypandoc
except:
    print("Make sure that you have pypandoc installed before running convert.", file=sys.stderr)
    exit(1)


class Convert(BaseCommand):
    """
    Using pypandoc, convert a file from one type of formatting to another.

    This can be used to convert from Markdown to HTML and back, as well as many
    other types of supported formats.

    The pypandoc Python package documentation can be found at:
        https://pypi.org/project/pypandoc/

    The pandoc System package documentation can be found at:
        https://pandoc.org/MANUAL.html
    """

    def __init__(self, options, *args, **kwargs):
        """
        Initialize the Convert command.

        :return {NoneType}:
        """
        super(Convert, self).__init__(options, depends_on_project=False, *args, **kwargs)
        self.supported_commands = "convert"

    def run(self):
        """
        Run the conversion tool (pandoc).

        Convert a file from one format into another, such as Markdown to HTML or
        HTML to Markdown.

        Example:
          - odoo_manager convert --source=my_module/readme.md --dest=my_module/static/description/index.html
            - Converts `my_module/readme.md` from Markdown to HTML and
              overwrites the result it to `my_module/static/description/index.html`.

          - odoo_manager convert --source=my_module/readme.md --to=html
            - Converts `my_module/readme.md` from Markdown to HTML and outputs
              the result to shell.

          - odoo_manager convert --source=my_module/readme.md --dest=my_module/static/description/index.html --append
            - Converts `my_module/readme.md` from Markdown to HTML and
              **appends** the result it to `my_module/static/description/index.html`.

        :return {NoneType}:
        """
        # Install pandoc, if requested
        if self.options.get("--install", False):
            pypandoc.download_pandoc()

        # Parse simple parameters
        append = self.options.get("--append", False)
        module_name = self.options.get("--module", False)
        verbose = self.options.get("--verbose", False)
        with_format = True if module_name else self.options.get("--with-format", False)

        # Parse more complicated parameters
        source, dest, from_type, to_type = self._parse_convert_options()

        # Build out pandoc method's options
        options = self._get_pandoc_python_options(source, dest, from_type, to_type, append)

        if verbose:
            shell.out("Running 'pypandoc.convert_file' with the following options:", color="yellow")
            pp(options, width=1)
            shell.out("")

        # Convert the file
        try:
            output = pypandoc.convert_file(**options)
        except OSError as error:
            shell.out(
                "Could not run the covert method. Try running convert with the --install flag or see the errors "
                " above/below... You can also try installing the pandoc system package manually: "
                "https://pypi.org/project/pypandoc/#installing-pandoc-manually",
                color="pink",
                is_error=True,
            )
            shell.out("{}".format(error), color="yellow", is_error=True)
            exit(1)

        # Format the file, if requested
        if dest and with_format:
            if to_type not in Format({}).supported_commands + Format({}).supported_extensions:
                shell.out("Formatting '{}' files is not supported".format(to_type), color="pink", is_error=True)
                exit(1)
            shell.run("odoo-manager format {} --source={}".format(to_type, dest))

        # Output the converted text if no destination was provided
        elif not dest:
            shell.out("Converted text...", color="yellow")
            shell.out("")
            shell.out(output)

    @staticmethod
    def _get_pandoc_python_options(source, dest, from_type, to_type, append=False):
        """
        Build the appropriate options for the "pypandoc.convert_file" Python
        command based on the provided parameters.

        See the PyPi package page or the source code for more information.
          - [PyPi](https://pypi.org/project/pypandoc/)
          - [Source](https://github.com/bebraw/pypandoc/blob/master/pypandoc/__init__.py#L104)

        :param source {str}: Path to the file containing text to be converted.
        :param dest {str}:
            (optional) Path to file to write/append converted text.
            If not provided, converted text will output to shell.
            Either the `dest` or `to_type` parameters must be provided.
        :param from_type {str}:
            (optional) The file type to be converted from.
            If not provided, pandoc will guess the type based on file extension.
            Examples: "markdown" or "html"
        :param to_type {str}:
            (optional) The file type to convert into.
            If not provided, pandoc will guess the type based on file extension.
            Either the `dest` or `to_type` parameters must be provided.
            Examples: "html" or "markdown"
        :param append {bool}:
            **NOT IMPLEMENTED**
            If False (by default), the `dest_path` file will be overwritten.
            If True, the description will be appended to the `dest_path` file.
        :return {str}: The "pandoc" conversion command.
        """
        # TODO: Implement append...
        if append:
            shell.out("Appending to existing file is not yet supported. Sorry!", color="yellow")
            exit(1)

        return {"source_file": source, "outputfile": dest or None, "format": from_type, "to": to_type}

    @staticmethod
    def _get_pandoc_shell_command(source, dest, from_type, to_type, append=False):
        """
        Build the appropriate "pandoc" shell command based on the provided
        parameters.

        See the [Pandoc Documentation](https://pandoc.org/MANUAL.html) for more
        details.

        ```
        # Command for system-level pandoc package
        cmd = "pandoc {options}".format(
            options=self._get_pandoc_shell_command(source, dest, from_type, to_type, append)
        )
        if verbose:
            shell.out(cmd)
        shell.run(cmd)
        ```

        :param source {str}: Path to the file containing text to be converted.
        :param dest {str}:
            (optional) Path to file to write/append converted text.
            If not provided, converted text will output to shell.
            Either the `dest` or `to_type` parameters must be provided.
        :param from_type {str}:
            (optional) The file type to be converted from.
            If not provided, pandoc will guess the type based on file extension.
            Examples: "markdown" or "html"
        :param to_type {str}:
            (optional) The file type to convert into.
            If not provided, pandoc will guess the type based on file extension.
            Either the `dest` or `to_type` parameters must be provided.
            Examples: "html" or "markdown"
        :param append {bool}:
            **NOT IMPLEMENTED**
            If False (by default), the `dest_path` file will be overwritten.
            If True, the description will be appended to the `dest_path` file.
        :return {str}: The "pandoc" conversion command.
        """
        # TODO: Implement append...
        if append:
            shell.out("Appending to existing file is not yet supported. Sorry!", color="yellow")
            exit(1)

        # Convert
        cmd = "pandoc"

        # From this type
        if from_type:
            cmd += " -f {}".format(from_type)

        # Into this type
        if to_type:
            cmd += " -t {}".format(to_type)

        # Using this as the destination file (otherwise output to shell)
        if dest:
            cmd += " -o {}".format(dest)

        # Using this as the source file
        cmd += " {}".format(source)

        return cmd

    def _get_destination_file(self, to_type):
        """
        Parse and validate the provided "--dest" and "--to" parameters to ensure
        the required options are available to pass to pandoc.

        If the "--module" parameter is provided, then the destination file will
        be "<module_name>/static/description/index.html" **unless** a path is
        provided via the "--dest" paramater.

        :param to_type {str}: The type of file being converted into.
        :return {str}: The path to the destination file, if provided.
        """
        paths = self.get_paths()
        module_name = self.options.get("--module", False)

        # Get the provided destination file path (if any)
        dest = self.options.get("--dest", False)

        if not dest:
            # If a module name is provided without a source file path, then use
            # `static/description/index.html` by default.
            if module_name:
                dest = "{}/static/description/index.html".format(module_name)
            # If module name is not provided and no "To" type is given, then
            # prompt for the destination path.
            elif not to_type:
                str(input(" > What is the path to the destination file? (Leave blank for output to shell) "))

        # If Destination is provided, then ensure it exists
        if dest and not os.path.isfile(paths.base(dest)):
            shell.out("The destination path ({}) does not appear to exist...".format(dest), color="pink", is_error=True)
            exit(1)

        # Destination file or "To" file type must be given for Pandoc to know
        # what to covert the source into.
        if not dest and not to_type:
            shell.out("You must provide a destination path or a file type to convert to", color="pink", is_error=True)
            exit(1)

        return dest

    def _get_source_file(self):
        """
        Parse and validate the provided "--source" parameter to ensure the
        required options are available to pass to pandoc.

        If the "--module" parameter is provided, then the source file will
        be "<module_name>/readme.md" **unless** a path is provided via the
        "--source" paramater.

        :return {str}: The path to the source file to be converted.
        """
        paths = self.get_paths()
        module_name = self.options.get("--module", False)

        # Get the provided source file path
        source = self.options.get("--source", False)

        if not source:
            # If a module name is provided without a source file path, then use
            # `readme.md` by default.
            if module_name:
                source = "{}/readme.md".format(module_name)
            # If module name is not provided, then prompt for the source path.
            else:
                str(input(" > What is the path to the source file? "))

        # Ensure source file exists
        if not source or not os.path.isfile(paths.base(source)):
            shell.out("The source path does not appear to exist...", color="pink", is_error=True)
            shell.out("    {}".format(source), color="yellow", is_error=True)
            exit(1)

        return source

    def _parse_convert_options(self):
        """
        Parse the options provided to the convert command to build the
        parameters which are required for pandoc.

        :return {tuple}:
            source {str}: The source file to be converted.
            dest {str | bool}:
                (optional) Path to file to write/append converted text.
            from_type {str}: The file type to be converted from.
            to_type {str}: The file type to convert into.
        """
        # Get "From" and "To" file types (if any)
        from_type = self.options.get("--from", False)
        to_type = self.options.get("--to", False)

        # Get "Source" and "Destination" filed
        source = self._get_source_file()
        dest = self._get_destination_file(to_type)

        # If the type is not provided, then get it from the destination file
        # extension. This is not required for shell command, but it is required
        # for the `convert_file` Python method.
        if not to_type:
            to_type = dest.split(".")[-1]

        return source, dest, from_type, to_type



import os


def check(params, errors):
    """
    Check if all of the parameters exist or not, and print the corresponding
    error if one is not passed through.

    :param params:
    :param errors:
    :return {NoneType}:
    """
    for index, param in enumerate(params):
        if not param:
            print(errors[index])
            exit(0)


def check_executable(ctx, exe):
    """
    Check that the system has a certain executable file command available to it
    before we try to start running commands that utilize it.

    :param ctx {invoke.context.Context}: Invoke context variable
    :param exe {str}:
    """
    current = ""

    try:
        for cmd in exe:
            current = cmd
            ctx.run("whereis {} > /dev/null".format(cmd))
    except Exception as e:
        ctx.run('echo "Make sure that you have the following packages installed properly: {}"'.format(", ".join(exe)))
        ctx.run('echo "It looks like you may have had a problem with the command: {}."'.format(current))
        exit(1)


def check_packages(ctx):
    """
    Check and make sure that all dependencies are configured to run the project
    in general. These are not project specific dependencies. Not talking about
    Odoo dependencies.

    This checks that the user has the proper packages available to them such as
    pip3, invoke, git, etc.

    :param ctx {invoke.context.Context}: Invoke context variable
    """
    check_executable(
        ctx,
        [
            "git",
            "docker",
            "docker-compose",
            "docker",
            "invoke",
            "curl",
            "echo",
            "sed",
            "pip3",
            "tail",
            "pylint",
            "yapf",
        ],
    )


def check_project(ctx):
    """
    Check that the project is ready to run basically. We need to make sure we
    have a few different things:

    - At least an odoo core instance.
    - An odoo log created.
    - An odoo conf file.

    :param ctx {invoke.context.Context}: Invoke context variable
    """

    must_have_files = [".env", ".container/config/odoo.conf", ".container/log/odoo.log", ".container/core/LICENSE"]

    for rel_filepath in must_have_files:
        path = os.path.abspath(rel_filepath)
        if not os.path.isfile(path):
            ctx.run('echo "Make sure that you have the following files configured:"')
            ctx.run('echo "{}"'.format(", ".join(must_have_files)))
            ctx.run('echo "It looks like you are missing {}."'.format(rel_filepath))
            return False

    with open(".env", "r") as f:
        if "enterprise=1" in f.read().lower():
            if not os.path.isfile(".container/enterprise/LICENSE"):
                ctx.run(
                    'echo "It looks like you have enterprise enabled. Make sure you have the source files setup in '
                    '.container/enterprise."'
                )
                return False

    return True


def check_env(config, required):
    """
    Check and make sure that the config (.env file) is setup properly.

    :param config {config.configparser.ConfigParser}:
    :param required {bool}:
    :return {bool}:
    """
    if not config or not config.sections():
        print(
            "Make sure you have the .env file and the .container/env/odoo/odoo.env file setup and that they have the [options] header."
        )
        exit(1)
    for param in required:
        if not config.has_option("options", param):
            print("Make sure that you have all of these required values in your .env:")
            for print_param in required:
                print("    - {}".format(print_param))
            exit(1)
    return True


def check_dirs():
    """
    Check and make sure that the required directories are setup properly based
    on the mapping define in the docker compose file.

    TODO: This is going to be used to ensure that everything is setup properly
    when a user tries to run the project or build the project.

    :return {bool}:
    """
    pass


def can_run(ctx):
    """
    Check and make sure that the project has everything that it needs to run
    before we actually try to run the project. This is meant to help prevent any
    confusion when someone is setting up a new project and running into errors.

    :param ctx {invoke.context.Context}: Invoke context variable
    :return {bool}:
    """
    return check_project(ctx)



from .base import BaseCommand


class Log(BaseCommand):
    """
    Represents the `odoo-manager log` command which tails the log file for the
    current project.
    """

    def run(self):
        """
        Runs the command. See the parent class for more details about how
        commands are organized and ran.

        :return {NoneType}:
        """
        grep = self.options.get("--grep")

        if grep:
            cmd = "tail -f {} | grep {}".format(self.paths.base(".container/log/odoo.log"), grep)
        else:
            cmd = "tail -f {}".format(self.paths.base(".container/log/odoo.log"))

        self.ctx.run(cmd, pty=True)



from invoke import Config, Context
from odoo_manager.core import configs
from odoo_manager.core import paths as path_helpers


class BaseCommand(object):
    def __init__(self, options, depends_on_project=True, *args, **kwargs):
        """
        Initialize a BaseCommand object.

        :param {dict} options: Command line arguments passed in.
        :param {bool} depends_on_project:
            Determines if the command requires an odoo project to exist in the
            current directory. Examples of this are command like `run`, which
            require a project to actually run. Command like `new` do not need
            an existing project because it creates one.
        """
        self.options = options
        self.supported_commands = ()
        self.args = args
        self.kwargs = kwargs
        self.ctx = Context(Config())
        self._paths = None
        if depends_on_project:
            self._paths = path_helpers.Paths()
            configs.init()

    @property
    def paths(self):
        """
        Abstraction for accessing paths for the current project.

        This depends on the `get_paths` method for caching and retreival of the
        Paths object.

        :return {odoo_manager.core.paths.Paths}:
            Returns a Paths object containing references to paths in the current
            project.
        """
        return self.get_paths()

    def get_paths(self):
        """
        Get the Paths object for the current project, and cache the object.

        :return {odoo_manager.core.paths.Paths}:
            Returns a Paths object containing references to paths in the current
            project.
        """
        if not self._paths:
            self._paths = path_helpers.Paths()
        return self._paths

    def run(self):
        """
        Implements general functionality for the run command. Adds support for
        subcommand via the `supported_commands` tuple.

        Each command class should call super when inheriting the run method.

        ```
        class MyCommand(BaseCommand):
            def run(self):
                subcommand = super(MyCommand, self).run()
                if not subcommand:
                    pass  # Perform some logic
        ```

        :return {NoneType}:
        """
        if not self.supported_commands:
            raise NotImplementedError("All commands must implement the run() method.")

        for command in self.supported_commands:
            if command in self.options and self.options[command]:
                method = "run_{}".format(command)
                if not hasattr(self, method):
                    raise NotImplementedError("The function {} has not been implemented yet.".format(method))
                getattr(self, method)()
                return method

        return None



import os
from .base import BaseCommand
from odoo_manager.core import shell, git
from odoo_manager.core.configs import get_project_manifest
from odoo_manager.core.misc import cleanup

YES_OPTIONS = ("y", "ye", "yes", "ya")


class Setup(BaseCommand):
    def __init__(self, options, *args, **kwargs):
        """
        Construct a Setup command object for `odoo-manager setup`.

        :return {NoneType}:
        """
        super(Setup, self).__init__(options, *args, **kwargs)
        self.supported_commands = ("dependencies", "mapping", "version")

    def run(self):
        """
        Runs when the `odoo-manager setup` command is run.

        :return {NoneType}:
        """
        subcommand = super(Setup, self).run()

        if not subcommand:
            print("setting up")

    def run_dependencies(self):
        """
        Runs when the `odoo-manager setup dependencies` command is run.

        :return {NoneType}:
        """
        options = {
            "--no-update": self.options.get("--no-update", False),
            "--only": self.options.get("--only", False),
            "--yes": self.options.get("--yes", False),
        }

        paths = self.get_paths()
        configuration = get_project_manifest()
        cleanup()

        if "depends" in configuration:
            effected_modules = []
            problem_modules = []
            dependencies = options["--only"].split(",") if options["--only"] else configuration["depends"]
            for dependency_name in dependencies:
                if not options["--no-update"] or (
                    options["--no-update"] and not os.path.isdir(paths.base("_lib/{}".format(dependency_name)))
                ):
                    effected_modules.append(dependency_name)
                    git.diff_dependency(self.ctx, dependency_name, configuration["depends"][dependency_name])

            if problem_modules:
                shell.out("The following modules had some problem ...", color="lightred")
                shell.out("  " + str.join("\n  ", problem_modules), color="lightgrey")

            if effected_modules:
                shell.out("")
                shell.out("Adding the following modules to _lib...", color="lightblue")
                shell.out("\n---\n", color="lightblue")
                shell.out("  " + str.join("\n  ", effected_modules), color="lightgrey")
                shell.out("\n---\n", color="lightblue")

                if (
                    options["--yes"]
                    or str(
                        input(
                            "Review any diffs above. Should we continue to overwrite? This will overwrite the _lib modules listed above with files "
                            "from the dependency source (y/n) :"
                        )
                    ).lower()
                    in YES_OPTIONS
                ):
                    for dependency_name in effected_modules:
                        git.update_dependency(self.ctx, dependency_name, configuration["depends"][dependency_name])
                shell.out("\nAll done.")
            else:
                shell.out("\nEverything up to date.")

        cleanup()



import os
from .base import BaseCommand
from odoo_manager.core import shell, git

try:
    import black
except:
    print("Make sure that you have black installed before running format.")
    exit(1)


class Format(BaseCommand):
    line_length = 120
    excludes = [
        "\.container",
        "\.git",
        "\.hg",
        "\.mypy_cache",
        "\.nox",
        "\.om",
        "\.tox",
        "\.venv",
        "__manifest__.py",
        "_build",
        "_lib",
        "_lib_static",
        "dist",
        "tasks",
    ]

    def __init__(self, options, *args, **kwargs):
        """
        Initialize the supported Formatter commands.

        :return {NoneType}:
        """
        super(Format, self).__init__(options, depends_on_project=False, *args, **kwargs)
        self.supported_commands = ("html", "python", "xml")
        self.supported_extensions = ("html", "py", "xml")

    def run_html(self):
        """
        Run the HTML formatter tool, xmllint, with appropriate options.

        Example:
          # Format the given HTML file
          odoo-manager format html --source=<path/to/my/file.html

        :return {NoneType}:
        """
        source = self.options.get("--source", False)
        if not source:
            shell.out("A source file must be provided.", color="pink", is_error=True)
        if os.path.isdir(source):
            shell.out("Formatting directories is not currently supported.", color="pink", is_error=True)

        try:
            self._run_xmllint(source)
        except Exception as error:
            shell.out(
                "Standard source file failed to format. Trying again with encoding and wrapper div...", color="yellow"
            )
            try:
                # It seems that xmllint requires (A) encoding information and
                # (B) an empty "wrapper" element in order to format.
                with open(source, "r") as read_file:
                    contents = read_file.read()
                with open(source, "w") as write_file:
                    write_file.write('<?xml version="1.0" encoding="UTF-8"?>\n<div>\n' + contents + "</div>\n")

                self._run_xmllint(source)
            except Exception as error:
                shell.out(
                    "Source file failed to format even with encoding and wrapper div", color="pink", is_error=True
                )
                raise error
            shell.out("Source file formatting completed successfully", color="green")

    def _get_black_options(self):
        """
        Format a set of options to pass to black.

        :return {str}: Formatted set of cli options for black.
        """
        return "-l {line_length} --exclude '{excludes}'".format(
            line_length=self.line_length,
            excludes="(__manifest__.py|__openerp__.py|/({})/)".format("|".join(self.excludes)),
        )

    def run_python(self):
        """
        Run the Python formatter tool, black, with appropriate options.

        Example:
          # Format all Python files in the current directory
          odoo-manager format python

          # Format only Python files in the given path
          odoo-manager format python --source=<path/to/my/<file.py>>

        :return {NoneType}:
        """
        cmd = "black {options} {source}".format(
            options=self._get_black_options(), source=self.options.get("--source", False) or "."
        )

        if self.options.get("--verbose", False):
            shell.out(cmd)

        shell.run(cmd)

    def run_xml(self):
        """
        Run the XML formatter tool, xmllint, with appropriate options.

        Example:
          # Format the given XML file
          odoo-manager format xml --source=<path/to/my/file.html

        :return {NoneType}:
        """
        source = self.options.get("--source", False)
        if not source:
            shell.out("A source file must be provided.", color="pink", is_error=True)
        if os.path.isdir(source):
            shell.out("Formatting directories is not currently supported.", color="pink", is_error=True)

        self._run_xmllint(source)

    @staticmethod
    def _get_xmllint_command(source):
        """
        Build the `xmllint` lint command with options for the given source file.

        This formats the source file, saving the output to a temporary file,
        then finally overwrites the source file with the temporary file. The
        temporary file is required because `xmllint` does not support formatting
        files in-place.

        :param source {str}: Relative path to file being formatted.
        :return {str}:
            The `xmllint` command to be run to format the given source file.
        """
        # XMLLINT_INDENT specifies 4 spaces for indentation instead 2 (default)
        return r"XMLLINT_INDENT=\ \ \ \  xmllint --format {source} --output {source}.tmp && mv {source}.tmp {source}".format(
            source=source
        )

    def _run_xmllint(self, source):
        """
        Build and run the `xmllint` command with options for the given source
        file.

        This is to be called by a wrapper method, like `run_html` and `run_xml`
        so that each method can perform any necessary changes to the source file
        before calling the format method.

        :param source {str}: Relative path to file being formatted.
        :return {NoneType}:
        """
        cmd = self._get_xmllint_command(source)

        if self.options.get("--verbose", False):
            shell.out(cmd)

        shell.run(cmd)



"""
odoo-manager.

Usage:
  odoo-manager --version
  odoo-manager -h | --help
  odoo-manager build [--no-cache] [--no-depends] [--no-setup] [--pull] [--verbose] [--user=<user>] [--password=<password>]
  odoo-manager convert [--module=<module_name>] [--source=<file>] [--dest=<file>] [--from=<file_type>] [--to=<file_type>] [--append] [--install] [--with-format] [--verbose]
  odoo-manager format (html|python|xml) [--source=<directory_or_file>] [--verbose]
  odoo-manager hello
  odoo-manager log [--grep=<criteria>]
  odoo-manager new (project|module) [--name=<project>] [-h | --help] [--verbose]
  odoo-manager ps [--all]
  odoo-manager run [--rebuild] [--update] [--detach] [--args=<build_args>] [--compose=<compose_file>] [--verbose] [--show-cmd]
  odoo-manager setup [dependencies|mapping|version] [--no-update] [--yes] [--only=<modules>] [--verbose]
  odoo-manager stop

Options:
  -a --all                          List all of the relevant resources.
  --append                          Append to file instead of overwriting.
  --args=<build_args>               Add flags to the odoo executable.
  -c --compose=<compose_file>       Reference to the docker compose file to use.
  -o --dest=<path>                  Path to file where converted text should be
                                    written/appended.
  -d --detach                       Detach the process after running.
  --grep=<criteria>                 Restrict the output by criteria.
  -h --help                         Show this screen.
  --install                         Try to install pandoc (for converting text).
  --from=<file_type>                Type of data converting from.
  -m --module=<module_name>            Pass a name for the module (to use default
                                    source and destination paths).
  --name=<project>                  Pass a name for the project.
  --no-update                       Only install new modules instead of updating
                                    existing.
  --only=<modules>                  Only install a certain set of modules
                                    instead of everything in the manifest.
  -r --rebuild                      Rebuild the container before running the
                                    command.
  --show-cmd                        Show the command that is executed.
  -s --source=<path>                Path to file (or directory) to be converted
                                    or formatted.
  --to=<file_type>                  Type of data converting to.
  -u --update                       Update the build before running the command.
  -v --verbose                      Enable verbose details.
  --version                         Show version.
  --with-format                     After conversion, run odoo-manager format.
  --yes                             Assume yes to all questions [default: False]

Examples:
  odoo-manager hello
  odoo-manager format python
  odoo-manager format python -s my_module
  odoo-manager format html --source=<my_module/static/description/index.html>
  odoo-manager convert -m my_module
  odoo-manager convert -s my_module/static/description/index.html -d my_module/readme.md
  odoo-manager convert -s my_module/readme.md -d my_module/static/description/index.html --with-format
"""

import traceback
from inspect import getmembers, isclass
from docopt import docopt

import odoo_manager
from odoo_manager.core import shell, configs
from odoo_manager.cli.exceptions import CommandException


def main():
    import odoo_manager.cli.commands

    options = docopt(__doc__, version=odoo_manager.version)

    # Here we'll try to dynamically match the command the user is trying to run
    # with a pre-defined command class we've already created.
    for (option_name, option_passed) in options.items():
        if hasattr(odoo_manager.cli.commands, option_name) and option_passed:
            module = getattr(odoo_manager.cli.commands, option_name)
            commands = getmembers(module, isclass)
            command = [command[1] for command in commands if command[0] != "BaseCommand"][0]
            command = command(options)

            try:
                command.run()

            # We will never show verbose information directly about a
            # CommandException because the user should not need to see the
            # trace on these. These are meant for clear breaking points
            # that we catch and display to the user. These are not unexpected
            # errors.
            #
            # For example, if a project directory cannot be found, this
            # exception will be raised. The user only needs to know that
            # a project directory cannot be found, but not the full stack
            # trace for this error.
            except CommandException as e:
                shell.out("{}\n".format(str(e)), color=shell.COLOR_RED)

            # These are unexpected error, and we will display trace information
            # to the user to provide them more details about what is going on,
            # if they have the --verbose flag enabled.
            except Exception as e:
                if options.get("--verbose", False):
                    shell.out(traceback.format_exc(), color=shell.COLOR_RED)
                else:
                    shell.out("{}\n".format(e), color=shell.COLOR_RED)



import odoo_manager
from subprocess import PIPE, Popen as popen
from unittest import TestCase


class TestHelp(TestCase):
    def test_returns_usage_information(self):
        output = popen(["odoo-manager", "-h"], stdout=PIPE).communicate()[0]
        self.assertTrue("Usage:" in output)

        output = popen(["odoo-manager", "--help"], stdout=PIPE).communicate()[0]
        self.assertTrue("Usage:" in output)


class TestVersion(TestCase):
    def test_returns_version_information(self):
        output = popen(["odoo-manager", "--version"], stdout=PIPE).communicate()[0]
        self.assertEqual(output.strip(), odoo_manager.version)



import os
import re
import getpass
from urllib.parse import quote
from .base import BaseCommand
from odoo_manager.core import shell, git


class New(BaseCommand):
    def __init__(self, options, *args, **kwargs):
        """
        Construct a New command object for `odoo-manager new`.

        :return {NoneType}:
        """
        super(New, self).__init__(options, depends_on_project=False, *args, **kwargs)
        self.supported_commands = ("project", "module")

    def run_project(self):
        """
        Handles `odoo-manager new project`.

        :return {NoneType}:
        """
        shell.out("Making a new project...")

        verbose = self.options.get("--verbose", False)
        name = self.options.get("--name", False)
        branch = "master"

        if not name:
            shell.out('Missing project name. Try "odoo-manager new project --name=my_project"')
            return

        try:
            shell.run(
                'git clone git@github.com:gobluestingray/odoo_project_scaffold.git {} --branch="{}" "{}"'.format(
                    "--quiet" if not verbose else "", branch, name
                )
            )
        except:
            username = quote(input("Username? "))
            password = quote(getpass.getpass("Password? "))

            try:
                shell.run(
                    'git clone https://{}:{}@github.com/gobluestingray/odoo_project_scaffold.git --quiet --branch="{}" "{}"'.format(
                        username, password, branch, name
                    )
                )
            except:
                shell.out("Invalid credentials.", color=shell.COLOR_RED)

        shell.run('rm -rf "./{}/.git"'.format(name))
        shell.out("Successfully created the project:")
        shell.out("  name:         {}".format(name), color="lightgrey")

    def run_module(self):
        """
        Handles `odoo-manager new module`.
        """
        verbose = self.options.get("--verbose", False)
        name = input(shell.out(" > What is the technical name of the module? ", color="lightblue", run=False))
        branch = input(
            shell.out(" > What is the version of the module (9.0, 10.0, 11.0, etc.)? ", color="lightblue", run=False)
        )

        module_display_name = input(
            shell.out(" > What is the display name of the module? ", color="lightblue", run=False)
        )
        module_category = input(shell.out(" > What is the category of the module? ", color="lightblue", run=False))
        module_tagline = input(
            shell.out(" > What is a tagline for the module (1 sentence description)? ", color="lightblue", run=False)
        )
        module_summary = input(
            shell.out(" > What is a summary/short explanation of the module? ", color="lightblue", run=False)
        )

        shell.out("\n")
        shell.out("Generating the module...")

        try:
            shell.run(
                'git clone git@github.com:gobluestingray/odoo_module_scaffold.git {} --branch="{}" "{}"'.format(
                    "--quiet" if not verbose else "", branch, name
                )
            )
        except:
            username = quote(input("Username? "))
            password = quote(getpass.getpass("Password? "))

            try:
                shell.run(
                    'git clone https://{}:{}@github.com/gobluestingray/odoo_module_scaffold.git --quiet --branch="{}" "{}"'.format(
                        username, password, branch, name
                    )
                )
            except:
                shell.out("Invalid credentials.", color=shell.COLOR_RED)

        shell.run('rm -rf "./{}/.git"'.format(name))

        # Wire the defaults to the module. There are a few things that every
        # module needs:
        #
        # 1. New readme.md file
        # 2. Updated __manifest__.py
        # 3. Updated static/description/index.html
        with open("./{}/readme.md".format(name), "w") as readme_file:
            readme_file.write("# {} ({})".format(module_display_name, name))

        with open("./{}/static/description/index.html".format(name), "w") as index_html_file:
            index_html_file.write(
                """
<section class="oe_container">
    <div class="oe_row oe_spaced">
        <div class="oe_span12">
            <h2 class="oe_slogan">{}</h2>
            <h3 class="oe_slogan">{}</h3>
        </div>
    </div>
</section>\n""".lstrip().format(
                    module_display_name, module_tagline
                )
            )

        manifest_content = False
        manifest_files = ("__manifest__.py", "__openerp__.py")

        for potential_manifest_file in manifest_files:
            if os.path.isfile("./{}/{}".format(name, potential_manifest_file)):
                with open("./{}/{}".format(name, potential_manifest_file), "r") as manifest_file:
                    manifest_content = manifest_file.read()
                    manifest_content = re.sub(
                        r'[\'"]+name[\'"]+:\s+[\'"]+(.*)[\'"]+,',
                        '"name": "{}",'.format(module_display_name),
                        manifest_content,
                    )
                    manifest_content = re.sub(
                        r'[\'"]+category[\'"]+:\s+[\'"]+(.*)[\'"]+,',
                        '"category": "{}",'.format(module_category),
                        manifest_content,
                    )
                    manifest_content = re.sub(
                        r'[\'"]+version[\'"]+:\s+[\'"]+(.*)[\'"]+,',
                        '"version": "{}",'.format(branch + ".0"),
                        manifest_content,
                    )
                    manifest_content = re.sub(
                        r'[\'"]+summary[\'"]+:\s+[\'"]+(.*)[\'"]+,',
                        '"summary": "{}",'.format(module_tagline),
                        manifest_content,
                    )
                    manifest_content = re.sub(
                        r'[\'"]+description[\'"]+:\s+[\'"]+(.*)[\'"]+,',
                        '"description": "{}",'.format(module_summary),
                        manifest_content,
                    )

                if manifest_content:
                    with open("./{}/{}".format(name, potential_manifest_file), "w") as manifest_file:
                        manifest_file.write(manifest_content)

        shell.out("Successfully created the module:")
        shell.out("  name:         {}".format(name), color="lightgrey")
        shell.out("  branch:       {}".format(branch), color="lightgrey")
        shell.out("  display name: {}".format(module_display_name), color="lightgrey")
        shell.out("  category:     {}".format(module_category), color="lightgrey")
        shell.out("  tagline:      {}".format(module_tagline), color="lightgrey")
        shell.out("  summary:      {}".format(module_summary), color="lightgrey")



from . import base
from . import build
from . import convert
from . import format
from . import hello
from . import log
from . import new
from . import ps
from . import run
from . import setup
from . import stop



import os
import sys
import filecmp
import getpass
from urllib.parse import quote

from . import configs
from . import paths as path_helpers
from odoo_manager.core import shell

pipe_dev_null = path_helpers.pipe_dev_null


def redirect(verbose=False):
    """
    :param verbose {bool}:
    :return {str}:
    """
    return "" if verbose else (" " + pipe_dev_null)


class console_colors(object):
    """
    Helper to hold some constant for shell colors.
    """

    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    BOLD = "\033[1m"
    ENDC = "\033[0m"


def print_diff_files(ctx, dcmp):
    """
    :param ctx {invoke.context.Context}: Invoke context variable
    :param dcmp:
    :return {NoneType}:
    """
    for name in dcmp.diff_files:
        print(
            console_colors.WARNING
            + "diff_file {} found in {} and {}".format(name, dcmp.left, dcmp.right)
            + console_colors.ENDC
        )
    for sub_dcmp in dcmp.subdirs.values():
        print_diff_files(ctx, sub_dcmp)


def _git_clone(ctx, url, branch, output, depth=1):
    """
    Does a system call on git clone, returns the exit status from that system.

    :param url {str}:
    :param branch {str}:
    :return {int}: Returns int representing success or failure
    """
    return ctx.run(
        "git clone '{url}' '{output}' --branch={branch} --depth={depth} --quiet{pipe}".format(
            url=url, branch=branch, output=output, depth=depth, pipe=redirect()
        )
    )


def _get_git_config(user="", password=""):
    """
    Get the username and password to be used for Git clone. By priority, this
    will use:

    :param user {str}: (optional) The username to be used for Git clone.
    :param password {str}: (optional) The password to be used for Git clone.
    :return {dict}:
        The Git configuration data to be used for cloning.
        Example:
            {
                "username": "my_github_user",
                "password": "mysecretpassword",
            }
    """
    git_config = {}
    if user and password:
        # Use the given user and password
        git_config["username"] = user
        git_config["password"] = password
    elif configs.config.has_option("options", "USERNAME") and configs.config.has_option("options", "PASSWORD"):
        # Use the configuration's user and password
        git_config["username"] = configs.config.get("options", "USERNAME")
        git_config["password"] = configs.config.get("options", "PASSWORD")
    else:
        # Request user input
        git_config["username"] = quote(input("Git username? "))
        git_config["password"] = quote(getpass.getpass("Git password? "))
    return git_config


def _get_git_urls(repo_url):
    if repo_url[:4] == "http":
        http_url = repo_url
        ssh_url = (
            repo_url.replace("https://", "git@")
            .replace("http://", "git@")
            .replace(".com/", ".com:")
            .replace(".org/", ".org:")
        )
    else:
        ssh_url = repo_url
        http_url = repo_url.replace("git@", "https://").replace(".com:", ".com/").replace(".org:", ".org/")

    return {"ssh": ssh_url, "http": http_url}


def _get_git_url_obscured(git_clone_url):
    """
    Remove the password from a Git clone URL.
    Example:
        Input:  git clone https://my_github_user:mysecretpassword@github.com/odoo/enterprise
        Output: git clone https://my_github_user:******@github.com/odoo/enterprise
    :param git_clone_url {str}: The Git URL which needs password to be obscured.
    :return {str}: The given URL with Git Password obscured, if found.
    """
    # Find the Git password, which would be between ":" and "@" characters.
    # https://stackoverflow.com/a/36247211/3330552
    password = re.search(r"https:\/\/[^:]*:([^@]*)@.*?$", git_clone_url).group(1)

    # If found, replace the password with asterisks
    return git_clone_url.replace(password, "******")


def git_clone(ctx, url, branch, repo_name, output, depth=1, user="", password=""):
    urls = _get_git_urls(url)
    try:
        shell.out("        *Trying to clone {}...".format(urls["ssh"]), color="yellow")
        _git_clone(ctx, urls["ssh"], branch, output, depth=depth)
    except:
        try:
            shell.out("        *Trying to clone {}...".format(urls["http"]), color="yellow")
            _git_clone(ctx, urls["http"], branch, output, depth=depth)
        except:
            try:
                credentials = _get_git_config(user, password)
                credentials_url = "https://{username}:{password}@{url}".format(
                    username=credentials.get("username"), password=credentials.get("password"), url=urls["http"][8:]
                )
                shell.out(
                    "        *Trying to clone {}...".format(_get_git_url_obscured(credentials_url)), color="yellow"
                )
                _git_clone(ctx, credentials_url, branch, output, depth=depth)
            except:
                shell.out("        Sorry, could not clone the repo {}".format(repo_name), color="red")


def diff_dependency_git(ctx, name, module_config, log=True):
    """
    :param name:
    :param module_config:
    :return: dict Paths for the modules
    """
    paths = path_helpers.Paths()

    if log:
        shell.out("  Processing {}...".format(name), color="yellow")

    repo_name = module_config["url"].replace(".git", "").split("/").pop()
    repo_temp_path = paths.base("_make_tmp/{}".format(repo_name))
    original_path = paths.base("_lib/{}".format(name))

    # If there is no repo cloned down into the temp folder yet, then we are going to git clone it
    # down so that we can access modules that we need to move into _lib.
    if not os.path.isdir(repo_temp_path):
        shell.out("    *Starting git clone for {}...".format(module_config["url"]), color="yellow")

        repo_url = module_config["url"]
        repo_branch = module_config["branch"]
        git_clone(ctx, repo_url, repo_branch, repo_name, output=paths.base("_make_tmp/{}".format(repo_name)))

    # Once we've got the modules and files setup, then we are going to check and see if we want
    # to run a diff on these these or not.
    if os.path.isdir(paths.base("{temp_path}/{module_name}".format(temp_path=repo_temp_path, module_name=name))):
        temp_module_path = paths.base("{temp_path}/{module_name}".format(temp_path=repo_temp_path, module_name=name))
        if log and os.path.isdir(original_path):
            print_diff_files(ctx, filecmp.dircmp(original_path, temp_module_path))
    else:
        shell.out(
            " ** Unable to find {module_name} in {repo}.".format(repo=module_config["url"], module_name=name),
            color="red",
        )
        exit(1)

    return {"original": original_path, "updated": temp_module_path}


def diff_dependency(ctx, name, module_config, log=True):
    """
    :param ctx {invoke.context.Context}: Invoke context variable
    :param name {str}:
    :param module_config:
    :return {NoneType}:
    """
    if "type" in module_config:
        fn = getattr(sys.modules[__name__], "diff_dependency_{type}".format(type=module_config["type"]))
        if fn:
            return fn(ctx, name, module_config, log)
    else:
        return diff_dependency_git(ctx, name, module_config, log)


def update_dependency(ctx, name, module_config):
    """
    :param ctx {invoke.context.Context}: Invoke context variable
    :param name {str}:
    :param module_config:
    :return {NoneType}:
    """
    diff_paths = diff_dependency(ctx, name, module_config, log=False)
    shell.out("  Updating lib with {}...".format(os.path.basename(diff_paths["original"])), color="yellow")
    os.system("rm -rf {}".format(diff_paths["original"]))
    os.system("cp -R {} {}".format(diff_paths["updated"], diff_paths["original"]))



import setuptools
import odoo_manager
from setuptools import Command, find_packages, setup
from subprocess import call


with open("readme.md", "r") as fh:
    long_description = fh.read()


class RunTests(Command):
    description = "run tests"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        errno = call(["py.test"])
        raise SystemExit(errno)


setuptools.setup(
    name="odoo_manager",
    version=odoo_manager.version,
    author="Blue Stingray",
    author_email="odoo@bluestingray.com",
    description="Odoo manager utility",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["docopt==0.6.2", "invoke==1.5.*", "configparser==3.7.3", "black>=20.*", "pypandoc==1.5"],
    packages=setuptools.find_packages(exclude=["docs", "tests*"]),
    python_requires=">=3.3",
    entry_points={"console_scripts": ["odoo-manager=odoo_manager.cli.cli:main"]},
    extras_require={"test": ["coverage", "pytest", "pytest-cov"]},
    cmdclass={"test": RunTests},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
    ],
)



from .base import BaseCommand
from invoke import UnexpectedExit
from odoo_manager.core import configs, shell, validation
from odoo_manager.core.git import redirect


class Run(BaseCommand):
    """
    Represents the `odoo-manager run` command which runs the current odoo
    project locally.
    """

    def run(self):
        """
        Runs the command. See the parent class for more details about how
        commands are organized and ran.

        :return {NoneType}:
        """
        validation.check_env(configs.config, ["PIP_EXE", "ODOO_EXE", "LOCAL_PORT"])

        if self.options.get("--rebuild", False):
            # Tasks.build(ctx, update=update, verbose=verbose)
            shell.out("Rebuild is not supported yet.", color="yellow")
            exit(0)

        args = self.options.get("--args", "")
        if (
            args
            and ("-i" in args or "--install" in args or "-u" in args or "--upgrade" in args)
            and ("-d" not in args and "--database" not in args)
        ):
            shell.out("Make sure you include the -d flag when trying to upgrade or install modules.", color="yellow")
            exit(1)

        compose = self.options.get("--compose", False) or "docker-compose.yml"
        verbose = self.options.get("--verbose", False)
        detach = self.options.get("--detach", False)

        # We are going to rewrite args every time no matter if it was passed
        # in or not. We don't want to run into a situation where a user uses
        # args once, and then has the same flags passed in over and over since
        # the odoo.env file will be written.
        #
        # If the user passes in args, then the next time does not, we should
        # assume we need to write an empty string to the odoo.env.
        if configs.config.has_option("options", "ODOO_FLAGS"):
            try:
                self.ctx.run("sed -i '' '/^ODOO_FLAGS/d' {}{}".format(configs.standard_env_path, redirect(verbose)))
            except:
                self.ctx.run("sed -i '/^ODOO_FLAGS/d' {}{}".format(configs.standard_env_path, redirect(verbose)))
        if configs.odoo_env_config.has_option("options", "ODOO_FLAGS"):
            try:
                self.ctx.run("sed -i '' '/^ODOO_FLAGS/d' {}{}".format(configs.odoo_env_path, redirect(verbose)))
            except:
                self.ctx.run("sed -i '/^ODOO_FLAGS/d' {}{}".format(configs.odoo_env_path, redirect(verbose)))

        shell.out("Running project...")
        if not verbose:
            shell.out(
                "You are not running in verbose mode, so container logs/output will not appear here. Tail "
                "logs or run this command with the --verbose flag to see output. It may take a minute for "
                "all containers to fully start.",
                color="yellow",
            )
        shell.out("  detach?: {}".format(detach), color="lightgrey")
        shell.out("  args:    {}".format(args or "None"), color="lightgrey")
        shell.out("  port:    {}".format(configs.config.get("options", "LOCAL_PORT")), color="lightgrey")
        shell.out(
            "  url:     http://localhost:{}".format(configs.config.get("options", "LOCAL_PORT")), color="lightgrey"
        )

        try:
            cmd = "docker-compose -f {} up {}{}".format(compose, "-d" if detach else "", redirect(verbose))
            if self.options.get("--show-cmd", False):
                shell.out(cmd)

            self.ctx.run("echo 'ODOO_FLAGS={}' >> {}".format(args or "", configs.odoo_env_path))
            self.ctx.run(cmd)
        except UnexpectedExit as e:
            # Handle the unexpected exit while trying to run docker-compose up.
            # There are a couple of scenarios. We only care about an error code
            # "1" coming back which means that our program died on it's own. We
            # do not care about the user stopping the command with a ctrl-c
            # input.
            if e.result.exited == 1:
                shell.out(
                    "There was a problem running docker-compose.{}".format(
                        " Try again with --verbose to see the error information." if not verbose else ""
                    ),
                    color="pink",
                )
